/*
	Prewitt plugin for AviSynth

	Copyright (C) 2006 Mohan
  
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.
	 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
		
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The author can be contacted at the doom9.org forum under mg262.
*/

#include "stdafx.h"
#include <cmath>

using namespace std;

void fillborder(PVideoFrame &frame)
{
	const unsigned char borderColor = 255;

	const int width = frame->GetRowSize();
	const int height = frame->GetHeight();
	const int pitch = frame->GetPitch();

	unsigned char *row = frame->GetWritePtr();

	for (int x = 0; x < width; ++x)  
		row[x] = borderColor;
	row += pitch; 

	for (int y = 1; y < height-1; ++y) 
	{  
		row[0] = borderColor;
		row[width-1] = borderColor;

		row += pitch;   
	}
	for (int x  = 0; x < width; ++x) 
		row[x] = borderColor;
	//row += pitch; 
}

class Prewitt : public GenericVideoFilter 
{  
	double multiplier;
public:
	Prewitt(PClip _child, double _multiplier, IScriptEnvironment* env);
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

Prewitt::Prewitt(PClip _child, double _multiplier, IScriptEnvironment* env) : GenericVideoFilter(_child), multiplier(_multiplier)
{
}

PVideoFrame __stdcall Prewitt::GetFrame(int n, IScriptEnvironment* env) 
{
	const int scaledmultiplier = static_cast<int>(multiplier*256);

	PVideoFrame source = child->GetFrame(n, env);
	PVideoFrame edge = env->NewVideoFrame(vi);

	const int width = source->GetRowSize();
	const int height = source->GetHeight();
	const int sourcepitch = source->GetPitch();
	const int edgepitch = edge->GetPitch();

	const unsigned char borderColor = 255;

	const unsigned char* sourcerow = source->GetReadPtr();
	unsigned char* edgerow = edge->GetWritePtr();

	for (int y = 0; y < height-2; ++y) 
	{
		for (int x  = 0; x < width-2; ++x) 
		{
			// 00 01 02
			// 10  12
			// 20 21 22
			int x00, x01, x02, x10, x12, x20, x21, x22;
			x00 = sourcerow[x + 0];
			x01 = sourcerow[x + 1];
			x02 = sourcerow[x + 2];

			x10 = sourcerow[x + 0 + sourcepitch];		
			x12 = sourcerow[x + 2 + sourcepitch];

			x20 = sourcerow[x + 0 + sourcepitch * 2];		
			x21 = sourcerow[x + 1 + sourcepitch * 2];		
			x22 = sourcerow[x + 2 + sourcepitch * 2];

			int p90  = x00 + x01 + x02       - x20 - x21 - x22;
			int p45  =    x01 + x02 - x10 + x12    - x20 - x21;//silly bug spotted by @kassandro
			int p180 = x00    - x02 + x10 - x12 + x20    - x22;
			int p135 = x00 + x01    + x10 - x12    - x21 - x22;

			int maxa = max( abs(p90), abs(p45));
			int maxb = max( abs(p180), abs(p135));
			int maxvalue = max(maxa, maxb);
			int multipliedmaxvalue = (maxvalue*scaledmultiplier)>>8;

			edgerow[x + 1 + edgepitch] = min(multipliedmaxvalue, 255);
		}
		sourcerow += sourcepitch;   
		edgerow += edgepitch;   
	}

	fillborder(edge);
	return edge;
}


class ThinPrewitt : public GenericVideoFilter 
{  
	double multiplier;
public:
	ThinPrewitt(PClip _child, double _multiplier, IScriptEnvironment* env);
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

ThinPrewitt::ThinPrewitt(PClip _child, double _multiplier, IScriptEnvironment* env) : GenericVideoFilter(_child), multiplier(_multiplier)
{
}

PVideoFrame __stdcall ThinPrewitt::GetFrame(int n, IScriptEnvironment* env) 
{
	//direction of gradient...
	const unsigned char uninitialised = 255;
	const unsigned char horizontal = 0;
	const unsigned char vertical = 1;
	const unsigned char downleft = 2;
	const unsigned char downright = 3;

	const int scaledmultiplier = static_cast<int>(multiplier*256);

	PVideoFrame source = child->GetFrame(n, env);
	PVideoFrame direction = env->NewVideoFrame(vi);
	PVideoFrame edge = env->NewVideoFrame(vi);

	const int width = source->GetRowSize();
	const int height = source->GetHeight();
	const int sourcepitch = source->GetPitch();
	const int directionpitch = direction->GetPitch();
	const int edgepitch = edge->GetPitch();

	const unsigned char borderColor = 255;

	const unsigned char* sourcerow = source->GetReadPtr();
	unsigned char* directionrow = direction->GetWritePtr();
	unsigned char* edgerow = edge->GetWritePtr();

	for (int y = 0; y < height-2; ++y) 
	{
		for (int x  = 0; x < width-2; ++x) 
		{
			// 00 01 02
			// 10  12
			// 20 21 22
			int x00, x01, x02, x10, x12, x20, x21, x22;
			x00 = sourcerow[x + 0];
			x01 = sourcerow[x + 1];
			x02 = sourcerow[x + 2];

			x10 = sourcerow[x + 0 + sourcepitch];		
			x12 = sourcerow[x + 2 + sourcepitch];

			x20 = sourcerow[x + 0 + sourcepitch * 2];		
			x21 = sourcerow[x + 1 + sourcepitch * 2];		
			x22 = sourcerow[x + 2 + sourcepitch * 2];

			int p90  = x00 + x01 + x02       - x20 - x21 - x22;
			int p45  =    x01 + x02 - x10 + x12    - x20 - x21;//silly bug spotted by @kassandro
			int p180 = x00    - x02 + x10 - x12 + x20    - x22;
			int p135 = x00 + x01    + x10 - x12    - x21 - x22;

				int maxvalue = 0;
				unsigned char localdirection = uninitialised;//unnecessary but avoids compiler warning
				if(abs(p45) >= maxvalue)
				{
					maxvalue = abs(p45);
					localdirection = downleft;
				}
				if(abs(p90) > maxvalue)
				{
					maxvalue = abs(p90);
					localdirection = vertical;
				}
				if(abs(p135) > maxvalue)
				{
					maxvalue = abs(p135);
					localdirection = downright;
				}
				if(abs(p180) > maxvalue)
				{
					maxvalue = abs(p180);
					localdirection = horizontal;
				}
				directionrow[x + 1 + directionpitch] = localdirection;
				int multipliedmaxvalue = (maxvalue*scaledmultiplier)>>8;
				edgerow[x + 1 + edgepitch] = min(multipliedmaxvalue, 255);
			}
		sourcerow += sourcepitch;   
		directionrow += directionpitch;   
		edgerow += edgepitch;   
	}
	//return direction;

	{
		int directionoffset[4];
		directionoffset[horizontal] = 1;
		directionoffset[vertical] = edgepitch;
		directionoffset[downleft] = edgepitch - 1;
		directionoffset[downright] = edgepitch + 1;


		PVideoFrame thinned = env->NewVideoFrame(vi);
		const int thinnedpitch = thinned->GetPitch();

		const unsigned char* directionrow = direction->GetReadPtr() + directionpitch;
		const unsigned char* edgerow = edge->GetReadPtr() + edgepitch;
		unsigned char* thinnedrow = thinned->GetWritePtr() + thinnedpitch;
		for (int y = 1; y < height-1; ++y) 
		{  
			for (int x  = 1; x < width-1; ++x) 
			{
				//if(directionrow[x]>=4) throw;
				int offset = directionoffset[  directionrow[x]  ];
				if(edgerow[x] < edgerow[x + offset] ||
					edgerow[x] < edgerow[x - offset])
					thinnedrow[x] = 0;
				else
					thinnedrow[x] = edgerow[x];
			}
			directionrow += directionpitch;   
			edgerow += edgepitch;   
			thinnedrow += thinnedpitch;   
		}

		fillborder(thinned);
		//copychroma(source, thinned);
		return thinned;
	}
}

AVSValue __cdecl Create_Prewitt(AVSValue args, void *user_data, IScriptEnvironment *env) 
{
	return new Prewitt(args[0].AsClip(), args[1].AsFloat(1.0), env); 
}

extern "C" __declspec(dllexport) const char *__stdcall AvisynthPluginInit2(IScriptEnvironment *env) 
{
	env->AddFunction("Prewitt", "c[multiplier]f", Create_Prewitt, 0);
	return "Prewitt plugin";
}
